#!/usr/bin/python

# Template for the 4th programming assignment.
# DB2 - WS 2015/16

# IMPORTS:

# Module to read linewise from standard input.
import fileinput

# Module to access command-line parameter and standard input.
import sys

# Module needed for ceil function.
import math

# IMPORTANT: In python indention matters. We follow Python's style guide 
# which uses 4 whitespace characters for an indention level.

# This is the main function.
def main():

    # Initialize a list of keys.
    keylist = []

    # Loop over each line from standard input.
    for line in fileinput.input("-"):

        # There is a newline at the end of each line - get rid of it.
        line_stripped = line.strip()

        # Convert each key explicitly to int.
        key = int(line_stripped)

        # Add a key to keylist.
        keylist.append(key)
        

    # Assign first command-line argument to variable nodedegree (number of 
    # pointers in a node).
    
    # NOTE: As in Java, the first element of an array (list in python)
    # is accessed via index 0. However, sys.argv[0] contains the name of python
    # script, not the first comand-line parameter.
    nodedegree = int(sys.argv[1])

    # Build the tree.
    nodelist, root_id = build_bptree(nodedegree, keylist)

    # Output - see README file for more details.
    #
    # Each line of the output represents a node (except the last line).
    # 
    # The first number is the node id.
    # Then the pointers and search keys alternate.
    # The values are separated by colons (":").
    # The last line outputs letter "r" followed by the ID of the root node.
    for node in nodelist:
        # Print node entries.
        # NOTE: str is necessary because join cannot join integers.
        print ":".join([ str(v) for v in node] )

    # Output ID of the root node.
    print "r%d" % root_id

# This function iterates to the leftmost leaf of childnode to retrieve its 
# leftmost key. 
def get_key(nodes, childnode):
    while True:
        if childnode[1] != -1:
            childnode = nodes[childnode[1]]
        else:
            return childnode[2]

# Function to build B+-Tree.
def build_bptree(nodedegree, keylist):

    # Sort all keys in keylist.
    keylist = sorted(keylist)

    # Initialise list of nodes which stores all nodes of the B+-Tree.
    # For more details see the README file.
    nodes = []

    # Compute how many leaf nodes are needed.
    # NOTE: in Python 2.x, division of integers returns an integer. However, 
    # we need to round up if the the number of keys is not a multiple of 
    # (nodedegree - 1). Therefore, we do the following steps:
    # - convert the number of keys to a float,
    # - divide,
    # - compute ceiling (round up),
    # - convert back to integer.
    nr_leafnodes = int(math.ceil(float(len(keylist)) / (nodedegree - 1)))

    # Partition keys into sub lists for leafnodes.
    # partitions is a list lists. Each element of partitions corresponds to
    # a single leaf node.
    partitions = []

    if nr_leafnodes != 1:

        # Iterate over leaf node partitions.
        for leafnode_nmb in range(nr_leafnodes):
            partitions.append(keylist[leafnode_nmb * (nodedegree - 1):(1 + leafnode_nmb) * (nodedegree - 1)])

    else:
        # Special case: only one node needed.
        partitions.append(keylist)


    # Store reference to last created leafnode (initially None).
    last_leafnode = None

    # Build all the leaf nodes.
    # Each leaf node is a list storing:
    # - node ID,
    # - pointers to child-nodes,
    # - keys.
    # For more details see README file.
    for partition in partitions:
        
        # Initialise new leaf node.
        leafnode = []

        # Node ID.
        node_nr = len(nodes)

        # Append node ID as the first element of the new leaf node.
        leafnode.append(node_nr)
        
        # Iterate over all keys in a leaf node.
        # Append pointer-key pairs to the new leaf node.
        for key in partition:

            # There is no pointer to children, append -1.
            leafnode.append(-1)

            # Append key.
            leafnode.append(key)

        # In the case when not all entries in a node are used, fill in 
        # the non-used entries with empty pointers and keys.
        for i in range(len(partition), nodedegree - 1):
            # There is no pointer to a child, append -1.
            leafnode.append(-1)

            # Empty key, put -1.
            leafnode.append( -1 )

        # Add pointer to the current leaf node in the previous leaf node.
        if last_leafnode != None:
            last_leafnode.append(node_nr)
        
        # If we are in the last node, we have to add the final pointer.
        if node_nr == nr_leafnodes - 1:
            leafnode.append(-1)

        # Add the new leaf node to the list of nodes.
        nodes.append(leafnode)

        # Make last_leafnode the current leafnode.
        last_leafnode = leafnode

    # If there is only one leaf node, we are done.
    if nr_leafnodes == 1:
        return nodes, node_nr

    # Copy nodes to lastlevelnodes list.
    lastlevelnodes = list(nodes)

    # Build inner nodes level by level bottom-up.
    while True:

        # Initialise list of nodes in the current level.
        levelnodes = []

        # Compute the number of nodes in the current level.
        # See comment for nr_leafnodes for details.
        nr_nodes_level = int(math.ceil(float(len(lastlevelnodes)) / (nodedegree)))

        # Initialise partitions.
        # Each element of partitions is a list of nodes from the lower level
        # which will be pointed to from a single node in the current level.
        partitions = []

        if nr_nodes_level != 1:
            # We are not in the root node.
            
            # Iterate over the nodes in the lower level.
            for node_level_nmb in range(nr_nodes_level):
                # Partition nodes of the lower level into sublists for building 
                # the nodes in the current level.
                partitions.append(lastlevelnodes[node_level_nmb * (nodedegree):(1 + node_level_nmb) * (nodedegree)])

        else:
            # We are in the root node.
            partitions.append(lastlevelnodes)

        # Build all nodes in the current level based on the partitions.
        for partition in partitions:
            
            # Initialise new node on the current level.
            node = []

            # Node ID.
            node_nr = len(nodes)

            # Append node ID as the first element of the current node.
            node.append(node_nr)
            
            # Append the first pointer which is the ID of the first node in 
            # the current partition (first child node).
            node.append(partition[0][0])

            # Iterate over the remaining children in a partition.
            for child_node in partition[1:]:

                # Append key (first key in the leftmost leaf node of child_node).
                node.append(get_key(nodes, child_node))
                
                # Append pointer to child_node.
                node.append(child_node[0])

            # If not all entries in the node are used, fill them in.
            for i in range(len(partition), nodedegree):
                
                # Empty key, put -1.
                node.append( -1 )

                # No pointer to a child, put -1.
                node.append(-1)

            # Add the current node to the list of all nodes and to the list of
            # current level nodes.
            nodes.append(node)
            levelnodes.append(node)

        # Copy nodes in the current level to lastlevelnodes list.
        lastlevelnodes = levelnodes

        # If we are in the root node.
        if nr_nodes_level == 1:
            # Return list of nodes and the ID of the root node.
            return (nodes, node_nr)

# The main function needs to be called explicitly.
if __name__ == "__main__":
    main()
